@@ -1,4 +1,7 @@
+from concurrent.futures import ThreadPoolExecutor
 import os
+from threading import Event
+import time
 from unittest import TestCase
 
 from mock import Mock
@@ -6,6 +9,8 @@ from mock import patch
 import pytest
 
 from dogpile.cache.region import _backend_loader
+from . import eq_
+from ._fixtures import _GenericBackendFixture
 from ._fixtures import _GenericBackendTest
 from ._fixtures import _GenericMutexTest
 
@@ -18,11 +23,11 @@ class _TestRedisConn(object):
     @classmethod
     def _check_backend_available(cls, backend):
         try:
-            client = backend._create_client()
-            client.set("x", "y")
+            backend._create_client()
+            backend.set("x", "y")
             # on py3k it appears to return b"y"
-            assert client.get("x").decode("ascii") == "y"
-            client.delete("x")
+            assert backend.get("x") == "y"
+            backend.delete("x")
         except Exception:
             if not expect_redis_running:
                 pytest.skip(
@@ -57,6 +62,68 @@ class RedisDistributedMutexTest(_TestRedisConn, _GenericMutexTest):
     }
 
 
+class RedisAsyncCreationTest(_TestRedisConn, _GenericBackendFixture, TestCase):
+    backend = "dogpile.cache.redis"
+    config_args = {
+        "arguments": {
+            "host": REDIS_HOST,
+            "port": REDIS_PORT,
+            "db": 0,
+            "distributed_lock": True,
+            # This is the important bit:
+            "thread_local_lock": False,
+        }
+    }
+
+    def test_distributed_async_locks(self):
+        pool = ThreadPoolExecutor(max_workers=1)
+        ev = Event()
+
+        # A simple example of how people may implement an async runner -
+        # plugged into a thread pool executor.
+        def asyncer(cache, key, creator, mutex):
+            def _call():
+                try:
+                    value = creator()
+                    cache.set(key, value)
+                finally:
+                    # If a thread-local lock is used here, this will fail
+                    # because generally the async calls run in a different
+                    # thread (that's the point of async creators).
+                    try:
+                        mutex.release()
+                    except Exception:
+                        pass
+                    else:
+                        ev.set()
+
+            return pool.submit(_call)
+
+        reg = self._region(
+            region_args={"async_creation_runner": asyncer},
+            config_args={"expiration_time": 0.1},
+        )
+
+        @reg.cache_on_arguments()
+        def blah(k):
+            return k * 2
+
+        # First call adds to the cache without calling the async creator.
+        eq_(blah("asd"), "asdasd")
+
+        # Wait long enough to cause the cached value to get stale.
+        time.sleep(0.3)
+
+        # This will trigger the async runner and return the stale value.
+        eq_(blah("asd"), "asdasd")
+
+        # Wait for the the async runner to finish or timeout. If the mutex
+        # release errored, then the event won't be set and we'll timeout.
+        # On <= Python 3.1, wait returned nothing. So check is_set after.
+        ev.wait(timeout=1.0)
+        eq_(ev.is_set(), True)
+
+
 @patch("redis.StrictRedis", autospec=True)
 class RedisConnectionTest(TestCase):
     backend = "dogpile.cache.redis"
